/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *
 * info@habanero.ncsa.uiuc.edu
 *
 * Copyright 1995-1996, Board of Trustees of the University of Illinois,
 * All Rights Reserved
 *
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 */
package ncsa.awt.draw;

import java.awt.*;
import java.util.*;
import java.io.*;
import ncsa.habanero.*;

/**
 * An abstract class that is subclassed for drawing operations
 *
 * @version	1.0 05/24/96
 * @author	NCSA SDG
 */
public abstract class DrawingOperation implements Marshallable {

	/** 
         * A list of drawing primitives
         */
         protected Vector plots = null;

        /**
         * A component that will be drawn into for this operation
         */
	protected Component canvas = null;

        /**
         * Indicates when to perform this action
         */
	protected int react = Event.MOUSE_DOWN;

        /**
         * The index in the plot vector of the primitive to perform
         * this action on.
         */
	protected int primitiveIndex = -1;

        /**
         * A point used as a reference for this operation
         */
	protected Point clickedPoint = null;

	/**
	 * Constructs a null drawing operation
	 */
	public DrawingOperation() {
	}

        /**
         * Initialize this drawing operation with a place to draw and
         * the primitives involved.
         * @param canvas the graphics context
         * @param plots the vector of drawing primitives
         */
	public void init(Component canvas, Vector plots) {
		this.canvas = canvas;
		this.plots = plots;
	}

	/** 
         * Sets the value of where the user clicked for this operation
         * @param x the x coordinate
         * @param y the y coordinate
         */
	public void setPoint(int x, int y) {
		clickedPoint = new Point(x,y);
	}

	/** 
         * Gets where the user clicked.
         * @return the Point which the user clicked
         */
	public final Point getPoint() {
		return clickedPoint;
	}

        /**
         * Gets the reaction value for this operation, which is used to
         * determine when the operation should be performed.
         * @return the reaction value (Event.MOUSE_DOWN, Event.MOUSE_UP)
         */
	public final int getReaction() {
		return react;
	}

        /**
         * Sets the index in the vector for the primitive we're interested in
         * @param index the primitive index
         */
	public void setIndex(int index) {
		primitiveIndex = index;
	}

        /**
         * Gets the index in the vector for the primitive we're interested in
         * @return the primtive index
         */
	public final int getIndex() {
		return primitiveIndex;
	}


	/**
         * Given a point value (x,y) locate the first primitive on top of the
         * stack that is located at this point.
         * @param x the x coordinate
         * @param y the y coordinate
         */
	public int locatePrimitive(int x, int y) {
		DrawingPrimitive primitive = null;

		if (plots == null)
			return -1;
			
		for (int i = plots.size()-1; i >= 0; i--) {
			primitive = (DrawingPrimitive)plots.elementAt(i);
			if (primitive.inside(x,y) == true) {
				setIndex(i);
				return i;
			}
		}
		return -1;
	}

        /**
         * Creates a new instance of this operation
         *
         * In order for a newInstance() method to  be called directly (without
         * this routine), the user would have to specify the actual name of
         * the object.  Ideally, this would just be by doing something like:
         * <pre>
         *      dp = (DrawingOperation)Class.forName("Grab").newInstance();
         * </pre>
         * However, since these routines exist within a package, the user
         * would have to specify a fully qualified package name and object
         * name to create the object.  It could end up looking like:
         * <pre>
         *      dp = (DrawingOperation)Class.forName("ncsa.drawing.Grab").newIns
tance();
         * </pre>
         * And, if the package ever changed locations, the code in the user
         * application would need to be modified.  In order to spare the user
         * this, I created this routine which handles everything internally.
         */
        public DrawingOperation newInstance() {
                DrawingOperation dp = null;
                String name = this.getClass().getName();
 
                try {
                        dp = (DrawingOperation)Class.forName(name).newInstance()
;
                } catch (Exception e) {
                        System.out.println(e);
                }
                return dp;
        }
		
	/** 
	 * Abstract method that must be implemented by the subclassed
	 * primitive.
	 */
	public abstract DrawingPrimitive operation();

	/**
         * Send out information about how to reconstruct this DrawingOperation
         * to DataOutput.
         * @param out the DataOutput to write to
         */
	public void MarshallSelf(DataOutput out) throws IOException {
		out.writeInt(primitiveIndex);
	}

	/**
         * Retrieve information about how to reconstruct this DrawingOperation
         * from DataInput
         * @param in the DataInput to read from
         */
	public void UnmarshallSelf(DataInput in) throws IOException {
		primitiveIndex = in.readInt();
	}
}
