/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *
 * info@habanero.ncsa.uiuc.edu
 *
 * Copyright 1995-1996, Board of Trustees of the University of Illinois,
 * All Rights Reserved
 *
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 */
package ncsa.awt.draw;

import java.awt.*;
import java.util.*;
import java.io.*;
import ncsa.habanero.*;

/**
 * An abstract class from which all other drawing primtives are derived.
 *
 * @version	1.0 05/24/96
 * @author	NCSA SDG
 */
public abstract class DrawingPrimitive implements Marshallable {
	protected Color color = null;
	protected Component canvas = null;
	protected boolean grabbed = false;
	protected Point grabbedAt = null;
	protected Rectangle bounding_box = null;
	protected Vector points = null;
	protected boolean enteringText = true;
	protected Key myKey = null;

	public DrawingPrimitive() {
	}

	/**
	 * Constructs an empty drawing primitive with a given bounding_box
         * @param rect the Rectangle representing a bounding box
	 */
	public DrawingPrimitive(Rectangle rect) {
		bounding_box = rect;
		points = new Vector();
		points.setElementAt(new Point(rect.x, rect.y), 0);
		points.setElementAt(new Point(rect.x+rect.width, rect.y+rect.height), 1);
	}

        /**
         * Constructs a DrawingPrimitive with a given set of points
         * @param pnts the Vector containing the list of points for this
         * drawing primitive
         */
	public DrawingPrimitive(Vector pnts) {
		points = pnts;
	}

        /**
	 * Initialize a bounding_box which contains this set of points
	 * @param pnts the Vector containing the DrawingPrimitive which are
         * used to construct the bounding box.
	 */
	public void setBoundingBox(Vector pnts) {
		Point p0 = (Point)pnts.elementAt(0);
		Point p1 = (Point)pnts.elementAt(1);
		bounding_box = new Rectangle(p0);
		bounding_box.add(p1);
	}

	/**
	 * Sets the component we're drawing into
	 * @param component the component
	 */
	public void setCanvas(Component canvas) {
		this.canvas = canvas;
	}

	/**
	 * Retrieves the Vector containing point data
	 */
        public final Vector getPoints() {
                return points;
        }

	/**
	 * Sets the Vector containing point data
	 * @param pnts the Point Vector
	 */
        public void setPoints(Vector pnts) {
                points = pnts;
        }
 
	/**
	 * Sets the color we're drawing with
	 * @param the Color
	 */
        public void setColor(Color color) {
                this.color = color;
        }

	/**
	 * Retrieves the color we're drawing with
	 * @return the color
	 */
        public Color getColor() {
                return color;
        }

	/**
	 * Sets up the vector of points we're going to do in this
	 * drawing operation, and calls the mouseDownAction with
	 * point coordinates.
	 * @param x the x coordinate when the mouseDown event occurred.
	 * @param y the y coordinate when the mouseDown event occurred.
	 */
	public void mouseDown(int x, int y) {
		points = new Vector();
		mouseDownAction(x, y);
	}

	/**
	 * Preforms the mouse dragging operation for this primitive.  If the
	 * primitive has been selected for a dragging operation, the move
	 * method is called.  Otherwise, it's a straight drawing operation,
	 * so the mouseDragAction method is called.
	 * @param x the x coordinate when the mouseDrag event occurred.
	 * @param y the y coordinate when the mouseDrag event occurred.
	 */
	public void mouseDrag(int x, int y) {
		if (grabbed) {
			move(grabbedAt, x,y);
		} else {
			mouseDragAction(x, y);
		}
	}

	/**
	 * Preforms the mouse up operation for this primitive.  If the
	 * primitive has been selected for a dragging operation, nothing
	 * is done.  Otherwise, it's a straight drawing operation,
	 * so the mouseUpAction method is called.
	 * @param x the x coordinate when the mouseUp event occurred.
	 * @param y the y coordinate when the mouseUp event occurred.
	 */
	public void mouseUp(int x, int y) {
		if (grabbed == false) {
			mouseUpAction(x, y);
		}
	}

	/**
	 * Returns whether or not this primitive has been grabbed
	 * @return a boolean value
	 */
	public final boolean grabbed() {
		return grabbed;
	}

	/**
	 * Grabs this primitive
	 * @param x the x coordinate where the primitive was grabbed
	 * @param y the y coordinate where the primitive was grabbed
	 */
	public void grab(int x, int y) {
		grabbedAt = new Point(x, y);
		grabbed = true;
		canvas.repaint();
	}	

	/**
	 * Releases the grab on this primitive
	 */
	public void release() {
		grabbed = false;
	}

	/**
	 * Redraws the region specified by bounding box
	 * @param x0 upper left hand x coordinate of bounding box
	 * @param y0 upper left hand y coordinate of bounding box
	 * @param x1 lower right hand x coordinate of bounding box
	 * @param y1 lower right hand y coordinate of bounding box
	 */
	public Rectangle redraw(int x0, int y0, int x1, int y1) {
 
		Rectangle rect = recalibrate(x0, y0, x1, y1);

		if (rect != null)  {
			// kludge the repaint area....this could be more precise
			canvas.repaint(rect.x-10, rect.y-10, rect.width+10, rect.height+10); 
		}
		
		return rect;
	}

	/**
	 * Given the specified bounding box, return a Rectangle that has
	 * been recalculated for the given points.  Java currently (JDK 1.0)
	 * doesn't allow negative height and width, so this routine is
	 * necessary recalibrate rectangular region specified.
	 * @param x0 upper left hand x coordinate of bounding box
	 * @param y0 upper left hand y coordinate of bounding box
	 * @param x1 lower right hand x coordinate of bounding box
	 * @param y1 lower right hand y coordinate of bounding box
	 */
	public Rectangle recalibrate(int x0, int y0, int x1, int y1) {
		Rectangle rect = new Rectangle(new Point(x0,y0));
		rect.add(x1,y1);

		return rect;
	}


	/**
 	 * Returns whether or not the given x,y coordinates are inside this
	 * primitive's region.
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public boolean inside(int x, int y) {
		if (bounding_box != null) {
			return bounding_box.inside(x,y);
		} 
			
		return false;
	}

	/**
	 * Move this primitive from the current location to a new (x,y)
         * @param start the reference Point at which this move started
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public void move(Point start, int x, int y) {
		if (bounding_box == null)
			return;

		int adjX = x - start.x;
		int adjY = y - start.y;

		Rectangle newArea = null;
		for (int i = 0; i < points.size(); i++) {
			Point pnt = (Point)points.elementAt(i);
			pnt.x = pnt.x + adjX;
			pnt.y = pnt.y + adjY;
			if (newArea == null)
				newArea = new Rectangle(pnt);
			else
				newArea.add(pnt);
		}

		start.x = x;
		start.y = y;

		Rectangle wholeArea = newArea.union(bounding_box);

		canvas.repaint(wholeArea.x, wholeArea.y, wholeArea.width+1, wholeArea.height+1);
		bounding_box = newArea;
	}

	/**
	 * Creates a new instance of this primitive.  
	 *
	 * In order for a newInstance() method to  be called directly (without
	 * this routine), the user would have to specify the actual name of
	 * the object.  Ideally, this would just be by doing something like:
	 *
	 *	dp = (DrawingPrimitive)Class.forName("Oval").newInstance();
	 *
	 * However, since these routines exist within a package, the user
	 * would have to specify a fully qualified package name and object
	 * name to create the object.  It could end up looking like:
	 *
	 *	dp = (DrawingPrimitive)Class.forName("ncsa.drawing.Oval").newInstance();
	 * 
	 * And, if the package ever changed locations, the code in the user
	 * application would need to be modified.  In order to spare the user
	 * this, I created this routine which handles everything internally.
	 */
	public DrawingPrimitive newInstance() {
		DrawingPrimitive dp = null;
		String name = this.getClass().getName();

		try {
			dp = (DrawingPrimitive)Class.forName(name).newInstance();
		} catch (Exception e) {
			System.out.println(e);
		}
		return dp;
	}

	/** 
	 * Initializes the data structures for this primitive on mouse down
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public void mouseDownAction(int x, int y) {
                points.addElement(new Point(x,y)); /* point 0 */
                points.addElement(new Point(x,y)); /* point 1 */
	}

	/**
         * An action that can be used by both a mouseDrag and mouseUp
	 * In this case, it updates the bounding box of the primitive
         * with the new (x, y) coordinate.  This is used for drawing and 
	 * DrawingPrimitive selection.
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public void commonAction(int x, int y) {
                Point p0 = (Point)points.elementAt(0);
                Point p1 = (Point)points.elementAt(1);
 
		Rectangle oldbox = new Rectangle(p0);
		oldbox.add(p1);

		Point newPoint = new Point(x,y);
                points.setElementAt(newPoint, 1);

		bounding_box = new Rectangle(p0);
		bounding_box.add(newPoint);

		bounding_box.width++;
		bounding_box.height++;

		oldbox.add(bounding_box);

		canvas.repaint(oldbox.x, oldbox.y, oldbox.width+1, oldbox.height+1);

	}

	/** 
	 * Updates the data structures for this primitive during a mouse drag
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public void mouseDragAction(int x, int y) {
		commonAction(x, y);
	}

	/** 
	 * Updates the data structures for this primitive during a mouse up
	 * @param x the x coordinate
	 * @param y the y coordinate
	 */
	public void mouseUpAction(int x, int y) {
		commonAction(x, y);
	}

	/**
	 * Performed when a key is pressed.  For most primitives, this is
         * empty.  It can be used to implement text input, or multi-action
         * drawing primitives.
         */
	public void keyDownAction(int key) {
	}

	/**
	 * Returns a boolean value to see if we're in a text input mode
	 * @return a boolean value
         */
	public final boolean enteringText() {
		return enteringText;
	}

	/**
	 * Draw the primitive in the given graphics context
	 * @param g the graphics context
         */
	public void drawPlot(Graphics g) {
		g.setColor(color);
		drawPoints(g);
		/* we could add more stuff (like highlighting) here */
	}

	/** 
	 * Abstract method that must be implemented by the subclassed
	 * primitive.
	 */
	public abstract void drawPoints(Graphics g);

	/**
	 * Sets a Habanero Key (used by a Habanero Lock) for this 
	 * DrawingPrimitive
         * @param key A Habanero Key
	 */
	public final void setKey(Key key) {
		myKey = key;
	}

	/**
	 * Retrieves a Habanero Key for this Primitive.
	 */
	public final Key getKey() {
		return myKey;
	}

	/**
	 * Marshalls this DrawingPrimitive on the given DataOutput
	 * @param out the DataOutput
	 */
	public void MarshallSelf(DataOutput out) throws IOException {

		out.writeInt(color.getRed());
		out.writeInt(color.getGreen());
		out.writeInt(color.getBlue());
		out.writeInt(points.size());
		for (int i = 0; i < points.size(); i++) {
			Point pnt = (Point)points.elementAt(i);
			out.writeInt(pnt.x);
			out.writeInt(pnt.y);
		}
	}

	/**
	 * UnMarshalls this DrawingPrimitive on the given DataInput
	 * @param out the DataInput
	 */
	public void UnmarshallSelf(DataInput in) throws IOException {

		int r = in.readInt();
		int g = in.readInt();
		int b = in.readInt();

		color = new Color(r,g,b);

		int size = in.readInt();
		points = new Vector();
		for (int i = 0; i < size; i++) {
			int x = in.readInt();
			int y = in.readInt();
			points.addElement(new Point(x,y));
		}
		setBoundingBox(points);
	}
}
