/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;
import java.io.*;
import java.net.InetAddress;
import ncsa.habanero.granted.Arbitrator;

/**
 * This action, executed on a client, creates a new Collobject.
 * It is generated by the server when it receives a NewCollobjectAction
 * and is the mechanism by which new applications are added to the server
*/
public class AddCollobjectAction extends Action {
  protected String collobname;
  protected String hostname;
  protected int portNum;
  protected Arbitrator parentArbitrator;

  /**
   *@param collobname - the name of the object which the new Collobject wraps
   *@param hostname - host name of the server
   *@param port - port number of the arbitrator the new Collobject's 
   *               arbitrator connects to
   *@param arb - parent Arbitrator of the new Collobject's arbitrator
   *@return Construct the collaboration object action
   */
  public AddCollobjectAction(String collobname, InetAddress address, 
			     int port, Arbitrator arb, 
			     String origin) {
    super();
    this.collobname = collobname;
    if (address != null) {
      byte[] addr = address.getAddress();
      this.hostname = (addr[0] & 0xFF) + "." + (addr[1] & 0xFF) + "." + 
	(addr[2] & 0xFF) + "." + (addr[3] & 0xFF);
    }
    else 
      hostname = "local";
    portNum = port;
    parentArbitrator = arb;
    originator = origin;  // The originator here is really the originator of
                          // the NewCollobjectAction
  }
  
  /**
   *@return Default constructor called if the next call will be a marshall()
   */
  public AddCollobjectAction() {
    super();
  }
  
  /**
   *@param target what this action will operate on
   *@return
   * This is the routine that makes the new Collobject and Wrapped objects.
   * A new DistributedArbitrator is created for the Collobject (this might
   * change to allow the clients to determine Arbitrator type) and the Wrapped 
   * class is instantiated from its name.  Finally, the Collobject is created.
   */
  public void act(Targetable target) {
    Sharable parent = (Sharable)target;
    DistributedArbitrator arb = 
      new DistributedArbitrator(parentArbitrator, hostname, portNum);
    parent.instantiate(collobname, arb, originator);
  }

  /**
   *@param out the output stream
   *@return
   * Marshall this by writing out the class name, host name, and port number,
   * and the specifier (starting at the root specifier) of the 
   * parent arbitrator.
   */
  public void MarshallSelf(DataOutput out) throws IOException {
    super.MarshallSelf(out);
    out.writeUTF(collobname);
    out.writeUTF(hostname);
    out.writeInt(portNum);
    parentArbitrator.MakeSpecifier(out, SessionManager.manager());
  }

  /**
   *@param in the input stream
   *@return
   * Unmarshall this by reading the class name, host name, and port number,
   * then find the parent arbitrator using its specifier.
   */
  public void UnmarshallSelf(DataInput in) throws IOException {
    super.UnmarshallSelf(in);
    collobname = in.readUTF();
    hostname = in.readUTF();
    portNum = in.readInt();
    parentArbitrator = (Arbitrator) SessionManager.manager().FindSpecifiedObject(in);
  }
  
}

