/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;
import java.util.*;
import java.io.*;
import java.net.*;
import ncsa.habanero.tickets.TicketGranter;
import ncsa.habanero.granted.*;


/**
 * This is the server-side portion of our centralized Arbitrator.  It
 * listens for connections from client-side Arbitrators, and once it 
 * gets them waits for ActionRequests.  Upon receiving an ActionRequest,
 * it decides if it is valid, and if so creates an ActionGranted which
 * it passes to its Router.
 */
public class CentralArbitrator extends Arbitrator implements Runnable {
  protected TicketGranter granter = new TicketGranter();
  protected Serializer serializer = new Serializer();  // if we want to log actions, they need to be serialized too

  protected Router router = null;
  int numParticipants = 0;
  ServerSocket server = null;
  int frozen = 0;
  boolean stopped = false;
  boolean recording = false;
  Recorder recordedActions = null;
  /**
   *@param granted the action that can be sent to everyone
   *@return
   * The extra indirection of this routine is provided if the 
   * router is not local to this object.  In that case, this
   * routine could be overridden to send the ActionGranted back
   * to the client, or to some other routing machine.
   */
  public synchronized void route_to_everyone(ActionGranted granted) {
    router().send_to_everyone(granted);
    Session s = session();
    if (s == null)
      return;  // not real sure how this can happen, but it has

    handleRecording(granted, s.isRecording());
  }

  protected void handleRecording(ActionGranted granted,
			    boolean sessionIsRecording) {
    if (recording != sessionIsRecording) {
      if (sessionIsRecording)
	recordedActions = new Recorder();
      else
 	recordedActions = null;
      recording = sessionIsRecording;
    }
    if (recording)
      recordedActions.addElement(granted);
  }
  
  public void resendMissedActions(int lastActionSeen) {
    System.out.println("I'm going to send actions after " + lastActionSeen);
    if (recordedActions != null) {
      int lastActionRecorded = recordedActions.lastElementNumber();
      System.out.println("lastActionRecorded is " + lastActionRecorded);
      while (lastActionSeen < lastActionRecorded) {
	System.out.println("rerouting Action #" + (lastActionSeen+1));
	route_to_everyone(recordedActions.elementAt(++lastActionSeen));
      }
    }
  }

  /**
   *@param parent the parent arbitrator
   *@param port the TCP port for the arbitrator to listen on
   *
   *@return
   * CentralArbitrator listens on a TCP port for incoming connections, adds
   * them to a list as they come in, and creates routers for each of them.
   * New connections each have a thread spawned off for them so that multiple
   * clients on those connections can speak to the CentralArbitrator 
   * concurrently.
   *
   * NOTE:  This does not handle race conditions.
   *
   */
   public CentralArbitrator(Arbitrator parent, Router r) {
     super(parent);
     init(parent, r);
   }
  
  /**
   *@return
   * Only use this if you are going to call init immmediately after
   */
  protected CentralArbitrator() {
     super(null);
   }

  /**
   *@param parent the parent arbitrator
   *@param r the router
   *@return sets up the arbitrator
   */
  public void init(Arbitrator parent, Router r) {
    SetSuperModel(parent);
    router = r;

    try {
       server = new ServerSocket(0);
//       System.out.println(getClass() + " is listening on port " + server.getLocalPort());
     } catch (IOException e) {
       System.out.println("Couldn't create socket");
//       System.exit(0);
// throw an exception instead
     }
     
     Thread thread = new Thread(this);
     thread.start();
  }
    
		     
  /**
   *@param out the output stream
   *@return
   * Write "important" info about this arbitrator to the output
   */
  public void writeInfo(DataOutput out) {
    try {
      out.writeInt(port());
    }
    catch (IOException e) {
      System.out.println("could not write info back");
      }
  }

  /**
   *@return increase the number of participants using this arbitrator
   */
  protected synchronized void increaseParticipants() {
    numParticipants++;
  }

  /**
   *@return decreases the number of participants using this arbitrator
   */
  protected synchronized void decreaseParticipants() {
    numParticipants--;
  }

  /**
   *@return
   * Wait for the server object to get a connection - when it does, add the
   * connection to our list.  Then tell the session a new connection was
   * made, and if the session OKs it, spawn a new thread to handle it.
   */
  public void run() {
     while (true) { /* run forever */
       try {
	 handleNewConnection(server.accept());
       }
       catch (IOException e) {
	   System.out.println("Accept failed!");
       }
     }
   }

  synchronized protected void handleNewConnection(Socket connection) {
    DataInputStream input = null;
    DataOutputStream output = null;
    try {
      input = new DataInputStream(connection.getInputStream());
      output = new DataOutputStream(connection.getOutputStream());
    } catch (IOException e) {
      System.out.println("Couldn't get an input stream from Socket " + 
			 connection.toString());
      
    }
	 
    try {
      String id = input.readUTF();
      if (session().newConnection(id, owner, output)) 
	new CentralRequestHandler(this, input, output, 
				  session().getParticipant(id));
    } catch (IOException e) {
      System.out.println("New connection didn't get a string");
      /* er, what do we do now? */
    }
  }


  /**
   *@param lkNew - the new lock that is requested
   *@return returns the key that allows you to use this lock
   */
  public Key getLock (Lock lkNew) {
    return getLock(lkNew, null);
  }

  /**
   *@param lkNew   the new lock that is requested
   *@param oldKey  the key that gives us permission to get this lock
   *@return A central arbitrator can't give anyone a key, since all
   *   actions are allowed
   */
  public Key getLock (Lock lkNew, Key oldKey) {
    return Key.NoKey;
  }

  /**
   *@param lkFinished   the key that we have finished using
   *@return does nothing, since a central arbitrator doesn't limit actions
   */
  public void releaseLock (Key lkFinished) {
  }

  /**
   *@param request the action which someone wants performed
   *@return
   * By default, the arbitrate routine merely calls a routine
   * which should ask the router to route the ActionGranted to
   * everyone.  This should be overridden by arbitrators which do real 
   * work.  However, if the action implements ArbitratorAction,
   * it takes over the routing responsibility instead.
   *@param request the ActionRequest we want to grant
   */
  public synchronized boolean arbitrate(ActionRequest request) {
    try {
      ArbitratorAction ac = (ArbitratorAction)request.action();
      if (ac.sendToEveryone())
	route_to_everyone(grant_request(request));
      ac.actAtArbitrator(this, request.target());
    }
    catch (ClassCastException e) {
      route_to_everyone(grant_request(request));
    }
    return true;
  }
  
  /**
   *@return This arbitrator's TicketGranter
   */
  public TicketGranter ticket_granter() {
    return granter;
  }

  /**
   *@return This arbitrator's Serializer
   */
  public Serializer serializer() {
    return serializer;
  }  

  /**
   *@return
   * If this arbitrator has its own router, return it.  Otherwise 
   * use the superclass' method to find it.
   */
  public Router router() {
    if (router != null)
      return router;
    else
      return super.router();
  }

  /**
   *@param input the input stream
   *@param output the output stream
   *@return
   * Once a connection has been made, listen for new ActionRequests.  When
   * we get one, unmarshall and arbitrate it.
   */
  public void listen(DataInput input, DataOutput output, 
		     SessionParticipant toWho) {
    increaseParticipants();
    while (true) {
      try {
	checkTemperature();
	Object request = Marshalling.Unmarshall(input);
	if (request == null)
	  continue;

	if (request instanceof ActionRequest) {
	  handleActionRequest((ActionRequest)request, output);
	}
	else {
	  handleQuestion((ArbitratorQuestion) request, output);
	}
      } catch (IOException e) {
	System.out.println("No longer listening");
	((TCPSessionParticipant)toWho).lostConnection(owner.MakeSpecifier(SessionManager.manager()),
						      this);
	decreaseParticipants();
	if (stopped && (numParticipants == 0)) {
	  cleanUp();
	}
	return;
      }
    }
  }


  protected synchronized void handleActionRequest(ActionRequest request, 
						  DataOutput output) throws IOException {
    boolean result = arbitrate(request);
    if (router == null) {
      output.writeBoolean(result);
    }
  }

  protected synchronized void handleQuestion(ArbitratorQuestion question,
					     DataOutput output) throws IOException {
    if (router == null) {
      if (question.method().equals("getLock"))
	getLock(question.descriptor(), question.key()).Marshall(output);
      else if (question.method().equals("blockingGetLock"))
	blockingGetLock(question.descriptor(), 
			question.key()).Marshall(output);
      else if (question.method().equals("releaseLock"))
	releaseLock(question.key());
    }
    else
      System.out.println("Cannot answer arbitration question");
  }
				     

  /**
   *@return closes the server
   */
  protected synchronized void cleanUp() {
    try {
      server.close();
    }
    catch (IOException e) {
      System.out.println("Could not cleanUp by closing the CentralArbitrator's ServerSocket");
    }
  }

  /**
   *@return The port I listen on for new connections.
   */
  public int port() {
    return server.getLocalPort();
  }

  /**
   *@param owner the new owner
   *@return
   * I set the sharable which is my owner.  If I have a router, set its
   * owner to the same thing.
   */
  public void set_owner(Sharable owner) {
    super.set_owner(owner);
    if (router != null)
      router.set_owner(owner);
  }

  /**
   *@param outStream the output stream
   *@return
   * myself by writing the port I listen on.
   */
  public void MarshallSelf(DataOutput outStream) {
    try {
      outStream.writeInt(port());
    }
    catch (IOException e) {
      System.out.println("could not write info back");
    }
  }

  /**
   *@return
   * This routine blocks while the variable frozen is > 0.
   */
  synchronized protected void checkTemperature() {
    while (frozen > 0) {
      try {
	wait();
      }
      catch (java.lang.InterruptedException e) {
	System.out.println("Thread interrupted in wait state");
      }
    }
  }

  /**
   *@return
   * This is used when dumping state to ensure the state doesn't change 
   * while we are dumping.  It simply flags that the router should be frozen.
   * The next input will not be handled until frozen is 0.
   */
  synchronized void freezeInput() {
    frozen += 1;
  }

  /**
   *@return
   * Decrements frozen by one and wakes up any waiting threads.  When frozen
   * hits 0, this object starts handling input again.
   */
  synchronized void thawInput() {
    if (frozen > 0)
      frozen -= 1;
    notifyAll();
  }

  /**
   *@return stops the CentralArbitrator from running
   */
  public synchronized void stop() {
    stopped = true;
    GetSuperModel().RemoveSubModel(this);
  }

  public Session session() {
    if (stopped)
      return null;
    else
      return super.session();
  }

}

/**
 * This is a helper class that allows the CentralArbitrator to listen 
 * to many inputs at once.  It starts itself in a new thread to listen
 * to a specific input.
 */
class CentralRequestHandler implements Runnable {
  DataInput input;
  DataOutput output;
  Thread thread = null;
  CentralArbitrator arbitrator = null;
  SessionParticipant toWho;

  /**
   *@param arb the central arbitrator
   *@param input the input stream
   *@param output the output stream
   *@return
   * Construct this with the input and the CentralArbitrator I'm working for.
   */
  public CentralRequestHandler(CentralArbitrator arb, DataInput input,
			       DataOutput output, SessionParticipant p) {
    arbitrator = arb;
    this.input = input;
    this.output = output;
    toWho = p;
    start();
  }
  
  /**
   *@return starts the thread
   */
  public void start () {
    if (thread == null) {
      thread = new Thread(this);
      thread.start();
    }
  }
  
  /**
   *@return what you want the thread to do
   */
  public void run() {
    arbitrator.listen(input, output, toWho);
    System.out.println("thread finished");
  }
}

  class Recorder  {
    Vector v = new Vector();

    Recorder() {
    }

    ActionGranted elementAt(int val) {
      if (v.size() > 0)
	return (ActionGranted) v.elementAt(val - firstElement().ticket().serialNumber());
      else
	return null;
    }

    void addElement(ActionGranted el) {
      if ((v.size() == 0) || 
	  (el.ticket().serialNumber() == firstElementNumber() + v.size())) {
	v.addElement(el);
	System.out.println("added Action #" + el.ticket().serialNumber());
      }
      else 
	System.out.println("didn't add Action #" + el.ticket().serialNumber());

   }

    int firstElementNumber() {
      if (v.size() > 0)
	return firstElement().ticket().serialNumber();
      else
	return 0;
    }

    int lastElementNumber() {
      return firstElementNumber() + v.size() - 1;
    }

    int size() {
      return v.size();
    }

    private ActionGranted firstElement() {
      return (ActionGranted)v.elementAt(0);
    }
}

