/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;
import java.io.*;
import java.net.*;
import java.util.StringTokenizer;

/**
 * This class acts as both a DataInput and a DataOutput.  Its main 
 * purpose is to defer actually making a connection until
 * something is actually written or read.  It is created with a host name and
 * port number, but creating the underlying socket is deferred until the first
 * read or write.  It can be created without a port, in which case reads or
 * writes will block until a port is specified.
 */
public class DeferredIO implements DataInput, DataOutput {
  protected DataOutputStream out = null;  // a SocketOutputStream
  protected DataInputStream in = null;  // a SocketInputStream
  protected String host;
  protected int port = 0;
  protected boolean wait = true;
  protected boolean goodConnection = true;

  /**
   * Set the port, and wake up sleeping threads.
   */
  public synchronized void setPort(int port) {
    this.port = port;
    notifyAll();
  }

  /**
   * Wake up sleeping threads.
   */
  public synchronized void wakeUp() {
    notifyAll();
  }

  /**
   * Return the host name
   */
  public String hostname() {
    return host;
  }

  /**
   * Return the port number.
   */
  public int portNumber() {
    return port;
  }

  public boolean isConnected() {
    return (out != null);
  }

  public synchronized void checkConnection() throws IOException {
    while (!isConnected() && !connect()) {
      try {
	if (waitForConnection() && !host.equals("local")) {
	  System.out.println("waiting for a connection on " + host);
	  wait();
	}
	else
	  throw new IOException("couldn't connect to " + host);
      }
      catch (InterruptedException e) {
      }
    }
  }

  public boolean waitForConnection() {
    return wait;
  }

  /**
   * Make the connection.  Try to create a socket to the given host and port.
   * The host might contain both numeric and string data, so use a 
   * StringTokenizer to check both.  When the socket is created, get a 
   * DataOutputStream and a DataInputStream from it.
  */
  public synchronized boolean connect() {
    if (host.equals("local"))
      return false;
    if (!isConnected()) {
      Socket sock = null;
	
      StringTokenizer tok = new StringTokenizer(host, "/");
      try {
	sock = new Socket(tok.nextToken(), port);
      }
      catch (UnknownHostException e) {
	if (tok.hasMoreTokens()) {
	  try {
	    sock = new Socket(tok.nextToken(), port);
	  }
	  catch (UnknownHostException ee) {
	    System.out.println("No such host as " + host);
	    return false;
	  }
	  catch (IOException ioe) {
	    System.out.println("No such host as " + host);
	    return false;
	  }
	}
	else {
	  System.out.println("No such host as " + host);
	  return false;
	}
      }
      catch (IOException e) {
	System.out.println("Couldn't connect to server!");
	return false;
      }
 
      try {
	out = new DataOutputStream(sock.getOutputStream());
	in = new DataInputStream(sock.getInputStream());
      } catch (IOException e) {
	System.out.println("Couldn't create datastream!");
	return false;
      }
    }
    return true;
  }
  /**
   * Construct this with a host name and port number
   */
  public DeferredIO(String host, int port) {
    this.host = host;
    this.port = port;
  }

  /**
   * Construct this with a host name only
   */
  public DeferredIO(String host) {
    this.host = host;
  }

  /**
   * Construct this with a host name and port number
   */
  public DeferredIO(String host, int port, boolean doWait) {
    this(host, port);
//    System.out.println("making a DeferredIO to " + host + ":" + port);
    wait = doWait;
  }

  public synchronized void stop() {
//    System.out.println("stopping a DeferredIO: " + this);
    if (out != null) {
      try {
	out.close();
      }
      catch (IOException e) {
	System.out.println("DeferredIO.stop(): Couldn't close the connection");
      }
      out = null;
      host = "local";
    }
  }

    
  /**
   * These are all routines which implement DataInput and DataOutput.
   * They all call connect() first if the connection has not yet been 
   * made.
   */
  public void write(int b) throws IOException {
    checkConnection();
    out.write(b);
  }

  public void write(byte b[]) throws IOException {
    checkConnection();
    out.write(b);
  }

  public void write(byte b[], int off, int len) throws IOException {
    checkConnection();
    out.write(b, off, len);
  }


  public void writeBoolean(boolean v) throws IOException {
    checkConnection();
    out.writeBoolean(v);
  }

  public void writeByte(int v) throws IOException {
    checkConnection();
    out.writeByte(v);
  }

  public void writeShort(int v) throws IOException {
    checkConnection();
    out.writeShort(v);
  }

  public void writeChar(int v) throws IOException {
    checkConnection();
    out.writeChar(v);
  }

  public void writeInt(int v) throws IOException {
    checkConnection();
    out.writeInt(v);
  }
  
  public void writeLong(long v) throws IOException {
    checkConnection();
    out.writeLong(v);
  }

  public void writeFloat(float v) throws IOException {
    checkConnection();
    out.writeFloat(v);
  }

  public void writeDouble(double v) throws IOException {
    checkConnection();
    out.writeDouble(v);
  }

  public void writeBytes(String s) throws IOException {
    checkConnection();
    out.writeBytes(s);
  }

  public void writeChars(String s) throws IOException {
    checkConnection();
    out.writeChars(s);
  }

  public void writeUTF(String str) throws IOException {
    checkConnection();
    out.writeUTF(str);
  }

  public void readFully(byte b[]) throws IOException {
    checkConnection();
    in.readFully(b);
  }

  public void readFully(byte b[], int off, int len) throws IOException {
    checkConnection();
    in.readFully(b, off, len);
  }

  public int skipBytes(int n) throws IOException {
    checkConnection();
    return in.skipBytes(n);
  }

  public boolean readBoolean() throws IOException {
    checkConnection();
    return in.readBoolean();
  }

  public byte readByte() throws IOException {
    checkConnection();
    return in.readByte();
  }

  public int readUnsignedByte() throws IOException {
    checkConnection();
    return in.readUnsignedByte();
  }

  public short readShort() throws IOException {
    checkConnection();
    return in.readShort();
  }
  
  public int readUnsignedShort() throws IOException {
    checkConnection();
    return in.readUnsignedShort();
  }
  
  public char readChar() throws IOException {
    checkConnection();
    return in.readChar();
  }
  
  public int readInt() throws IOException {
    checkConnection();
    return in.readInt();
  }

  public long readLong() throws IOException {
    checkConnection();
    return in.readLong();
  }
  
  public float readFloat() throws IOException {
    checkConnection();
    return in.readFloat();
  }
  
  public double readDouble() throws IOException {
    checkConnection();
    return in.readDouble();
  }
  
  public String readLine() throws IOException {
    checkConnection();
    return in.readLine();
  }
  
  public String readUTF() throws IOException {
    checkConnection();
    return in.readUTF();
  }
}



