/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;
import java.io.*;
import java.net.*;
import ncsa.habanero.granted.*;
import ncsa.habanero.tickets.TicketGranter;
import java.util.Observer;
import java.util.Observable;
/**
 * This is the client-side portion of our centralized Arbitrator.
 * It has a connection to the server object (through a DeferredIO 
 * object), which does the actual work.  It can also get messages 
 * back fromthe server object, which might ask it to tell somebody that an
 * ActionRequest was refused, or to route the ActionGranted itself.
 * Presumably subclasses would be built to do these things.
 */
public class DistributedArbitrator extends Arbitrator {
  protected DeferredIO talker;
  protected Serializer serializer;  
  protected Router router;

  /**
   * The constructor takes, as arguments, the parent Arbitrator,
   * the host name and port number of the Arbitrator to connect
   * to, and a boolean indicating whether this Arbitrator has its
   * own router.
   */
  public DistributedArbitrator(Arbitrator mama, String host, 
			       int port, boolean doRouting) {
    super(mama);
    serializer = new Serializer(this);  
    talker = new DistributedArbitratorIO(host, port, session());
    if (doRouting)
      router = newRouter();
    talker.connect();
  }
  
  /**
   * This constructor is like the previous, except it defaults to
   * no router.
   */
  public DistributedArbitrator(Arbitrator mama, String host, 
			       int port) {
    this(mama, host, port, false);
  }

  /**
   * This is the default constructor, used for unmarshalling.
   */
  public DistributedArbitrator() {
    super(null);
  }

  /**
   * Marshall myself.  Write out the specifier of my parent, then
   * marshall my Serializer, write out the host name and port number
   * of my CentralArbitrator, and whether or not I have my own
   * Router.
   */  
  public void MarshallSelf(DataOutput out) throws IOException {
    GetSuperModel().MakeSpecifier(out, SessionManager.manager());
    serializer.Marshall(out);
    out.writeUTF(talker.hostname());
    out.writeInt(talker.portNumber());
    out.writeBoolean(router != null);
  }

  /**
   * Unmarshalling can be tricky, because this object could be created
   * before its parent is added into the hierarchy.  For now, we assume
   * the parent will take care of the problem after it is initialized.
   * Aside from that, we just recreate the object using the marshalled
   * information.
   */  
  public void UnmarshallSelf(DataInput in) throws IOException {
    ModelObject parent = SessionManager.manager().FindSpecifiedObject(in);
    // HACK - for now we assume this will get fixed later
    if (parent != this)
      SetSuperModel(parent);
    serializer = (Serializer)Marshalling.Unmarshall(in, this);
    talker = new DistributedArbitratorIO(in.readUTF(),
					 in.readInt(),
					 session());
    if (in.readBoolean()) {
      router = newRouter();
      router.freezeInput();
    }
    talker.connect();
  }

  /**
   * The arbitrate routine sends the request to the CentralArbitrator. 
   *@param request the ActionRequest we want to grant
   */
  public synchronized boolean arbitrate(ActionRequest request) {
    if (session().isShared()) {
      try {
	request.Marshall(talker);
	if (router == null)
	  return talker.readBoolean();
	else
	  return true;
      }
      catch (IOException iox) {
	return false;
      }
    }
    else {
      request.sendDirectly(this);
      return true;
    }
  }
  
  /**
   * This object has no TicketGranter
   */  
  public TicketGranter ticket_granter() {
    return null;
  }
  
  /**
   * Return this object's serializer.
   */  
  public Serializer serializer() {
    return serializer;
  }  

  /**
   * Create and return a new DistributedRouter which routes
   * through the same socket as the talker.
   */  
  public Router newRouter() {
    return new DistributedRouter(owner, talker);
  }

  /**
   * If I have a router, return it.  Otherwise call up to my
   * superclass.
   */  
  public Router router() {
    if (router != null)
      return router;
    else
      return super.router();
  }


  /**
   * Return the hostname of the machine my CentralArbitrator 
   * is on.
   */  
  public String serverHost() {
    return talker.hostname();
  }

  /**
   *
   *  getLock
   *
   *  Requires:  lkdNew
   *  Modifies:  itself
   *  Effects:   returns a key to a locker if it is legal to do so
   *				is OK to do so
   *
   */

  public Key getLock (Lock lkdNew) {
    return getLock(lkdNew, null);
  }
  

  protected void askQuestion (String question, Lock lkdNew, Key oldKey) {
    if (!isDistributed())
      return;

    if (router == null) {
      ArbitratorQuestion aq = 
	new ArbitratorQuestion(question, oldKey, lkdNew);
      try {
	aq.Marshall(talker);
      }
      catch (IOException e) {
	System.out.println("Could not marshall a " + question + " request");
      }
    }
    else {
      System.out.println("Could not " + question + " because socket is in use");
    }
  }

  public Key getLock (Lock lkdNew, Key oldKey) {
    if (!isDistributed())
      return null;

    if (router == null) {
      askQuestion("getLock", lkdNew, oldKey);
      try {
	return (Key) Marshalling.Unmarshall(talker);
      }
      catch (IOException e) {
	System.out.println("Could not unmarshall a getLock request");
	return null;
      }
    }
    else {
      System.out.println("Could not get a key because socket is in use");
      return null;
    }
  }

  public Key blockingGetLock (Lock lkdNew, Key oldKey) {
    if (!isDistributed())
      return null;

    if (router == null) {
      askQuestion("blockingGetLock", lkdNew, oldKey);
      try {
	return (Key) Marshalling.Unmarshall(talker);
      }
      catch (IOException e) {
	System.out.println("Could not unmarshall a getLock request");
	return null;
      }
    }
    else {
      System.out.println("Could not get a key because socket is in use");
      return null;
    }
  }

    /**
     *
     *  releaseLock
     *
     *  Requires:  lkFinished - the key of the lock that is no longer needed
     *  Modifies:  itself
     *  Effects:   releases the lock
	 *
     */

  public void releaseLock (Key lkFinished) {
    if (router == null) {
	askQuestion("releaseLock", null, lkFinished);
    }
    else {
      System.out.println("Could not release a lock because socket is in use");
    }
  }

  public synchronized void stop() {
    talker.stop();
    if (router != null)
      router.stop();
    GetSuperModel().RemoveSubModel(this);
  }

  public boolean isDistributed() {
    return talker.isConnected();
  }

  /**
   * Set the Sharable which owns this object.  This is used to rectify a
   * chicken-egg problem when constructing ModelObjects and Arbitrators.
   */
  public synchronized void set_owner(Sharable owner) {
    super.set_owner(owner);
    if (router != null)
      router.set_owner(owner);
  }


}

/**
 * This is a subclass of DeferredIO which, after connecting,
 * an id which uniquely determines the participant using it.
 */
class DistributedArbitratorIO extends DeferredIO implements Observer {
  Session session;

  public DistributedArbitratorIO(String host, int port, Session s) {
    super(host, port);
    session = s;
  }

  public synchronized boolean connect() {
    if (out == null) {
      if (super.connect()) {
	try {
	  out.writeUTF(SessionManager.manager().id());
	}
	catch (IOException e) {
	  System.out.println("could not write id after connecting");
	  return false;
	}
	if (session != null)
	  session.addArbitratorConnection(this);
      }
      else
	return false;
    }
    return true;
  }

  public void update(Observable o, Object arg) {
    if (arg == session) {
      session = null;
      stop();
    }
    else
      System.out.println("Somehow DIO.update was called for the wrong session: " + arg);
  }

  public synchronized void stop() {
    super.stop();
    if (session != null)
      session.removeArbitratorConnection(this);
  }
}
