/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 */
package ncsa.habanero;
import ncsa.habanero.*;
import ncsa.habanero.granted.*;
import ncsa.habanero.tickets.*;
import java.io.DataOutput;
import java.io.DataInput;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration;

/**
 *
 *  A LimitArbitrator is an arbitrator that forces you to have a lock
 *  for all actions except the quit action.  The LimitArbitrator 
 *  provides keys, compares a new lock against older locks, and
 *  requires all actions to have keys.  Also allows actions or
 *  keys to block until allowed.
 *
 */
public class LimitArbitrator extends CentralArbitrator {

  //  These are the valid locks
  protected Vector validLockSet = new Vector();

  /**
   *@param parent the parent arbitrator
   *@param r the router
   *@return a new arbitrator
  */
  protected LimitArbitrator(Arbitrator parent, Router r) 
  {
    super(parent, r);
  }

  /**
   *@return creates a new arbitrator
  */
  protected LimitArbitrator() 
  {
    super();
  }

  protected boolean doArbitration(ActionRequest request) {
    boolean bResult = false;

    Key aKey = request.getKey ();
    if ((aKey == null) || aKey.equals(Key.NoKey)) {
      //  Check to see if this is an action that can go through
      if (!keylessAction (request.action()))
	return false;
	  
      //  Check to see if anyone blocks this
      if (anyBlocks (request.action()))
	return false;

      //  Go ahead with the action
      return super.arbitrate (request);
    }

    //  Use the lock
    int nUse = validLockSet.indexOf (aKey);
    Locker toUse = (Locker) validLockSet.elementAt (nUse);

    if (toUse == null) {
      //System.out.println("LimitArbitrator::arbitrate -- no such lock");
      return false;
    }

    if (toUse.use(request)) {
      //System.out.println("LimitArbitrator::arbitrate -- using the lock");
      bResult = super.arbitrate (request);
    }

    if (!toUse.valid()) {
      releaseLock (nUse);
    }

    return bResult;
  }
  
  /**
   * These routines should be overridden to do something with the
   * request.  The result should be an ActionGranted sent to the
   * router.
   *@param   request the action request to be performed
   *@return  true if the action was performed, false if it was not able to
  */
  public synchronized boolean arbitrate(ActionRequest request) {
    boolean result = doArbitration(request);
    while (request.blocks() && !result) {
      System.out.println("LimitArbitrator.arbitrate() waiting");
      try {
	wait();
      }
      catch (InterruptedException e) {
	System.out.println("LimitArbitrator.arbitrate() waking up");
      }
      result = doArbitration(request);
    }
   return result;   
  }

  /**
   *
   *@param   ldNew - the new lock descriptor
   *@param   oldKey - the key used to acquire this new lock 
   *@return  returns a key for a lock built from the lock 
   *              descriptor if it is OK to do so
   *
   */
  public Key getLock (Lock ldNew, Key oldKey) {
    if (!isDistributed())
      return null;

    if ((ldNew == null) || anyConflicts (ldNew, oldKey)) {
      	return Key.NoKey;
      }
    else {
      	Locker lkResult = new Locker (ldNew);
      	validLockSet.addElement (lkResult);
      	return lkResult.getKey ();
    	}
  }

  public synchronized Key blockingGetLock (Lock ldNew, Key oldKey) {
    if (!isDistributed())
      return null;

    if (ldNew == null)
      return Key.NoKey;
    Key result = getLock(ldNew, oldKey);
    while (result.equals(Key.NoKey)) {
      try {
	System.out.println("LimitArbitrator.blockingGetLock() waiting for key");
	wait();
      }
      catch (InterruptedException e) {
	System.out.println("LimitArbitrator.blockingGetLock() waking up");
	result = getLock(ldNew, oldKey);
      }
    }
    return result;
  }
  
  /**
   *
   *@param   kFinished - the key of the lock that is no longer needed
   *@return  releases the lock
   *
   */
  public synchronized void releaseLock (Key kFinished) {
    validLockSet.removeElement(kFinished);
    notify();
  }

  /**
   *
   *@param   kFinished - the index of the lock that is no longer needed
   *@return  releases the lock
   *
   */
  protected synchronized void releaseLock (int kFinished) {
    validLockSet.removeElementAt(kFinished);
    notify();
  }

  /**
   *
   *@param   kUse - the key we are going to use
   *@return  determines if the lock is good, and returns 
   *             true if it has been used
   *
   */
  public boolean validate (Key kUse) {
    return (kUse.valid() && (findLocker(kUse) !=  null));
  }

  
  /**
   *
   *@param   ldRequest - the description the user is requesting
   *@return  returns true if there is a conflicts with earlier locks
   *
   */

  protected boolean anyConflicts (Lock ldRequest, Key oldKey) 
    {
      //System.out.println ("LimitArbitrator::anyConflicts");

      int ndx = 0;
      int nLockCount = validLockSet.size ();
      
      while (ndx < nLockCount) {
      	Locker aLocker = (Locker) validLockSet.elementAt(ndx);
	
      	if (aLocker.valid()) {
	  //System.out.println ("LimitArbitrator::anyConflicts -- Valid Lock");

	  if (aLocker.conflict(ldRequest, oldKey)) {
	    //System.out.println ("LimitArbitrator::anyConflicts -- Conflict!!!");
	    return true;
	  }
	  ndx++;
	}
      	else {
	  releaseLock(ndx);
	}
      }
      
      return false;
    }
   

  /**
   *
   *@param   act - the action that we are checking to see if anyone objects
   *@return  returns true if a lock objects to the action
   *
   */

  protected boolean keylessAction (Action act) 
    {
      //System.out.println ("LimitArbitrator::keylessAction");

      if (act instanceof QuittingAction)
	return true;
      
      return false;
    }
   
  
  /**
   *
   *@param   act - the action that we are checking to see if anyone objects
   *@return  returns true if a lock objects to the action
   *
   */

  protected boolean anyBlocks (Action act) 
    {
      //System.out.println ("LimitArbitrator::anyBlocks");

      int ndx = 0;
      int nLockCount = validLockSet.size ();
      
      while (ndx < nLockCount) {
      	Locker aLocker = (Locker) validLockSet.elementAt(ndx);
	
      	if (aLocker.valid()) {
	  //System.out.println ("LimitArbitrator::anyBlocks -- Valid Lock");

	  if (aLocker.blocks (act)) {
	    //System.out.println ("LimitArbitrator::anyBlocks -- Blocked this X@!#@&!");
	    return true;
	  }
	  ndx++;
	}
      	else {
	  releaseLock(ndx);
	}
      }
      
      return false;
    }
   
  /**
   *
   *@param   kSearch - the key of the lock we are searching for
   *@return  returns the lock if it exists, and null otherwise
   *
   */
  protected Locker findLocker (Key key) {
    int nIndex = validLockSet.indexOf (key);
    if (nIndex == -1)
      return null;
    else
      return (Locker) validLockSet.elementAt (nIndex);
  }

}

