/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;

import java.io.DataOutput;
import java.io.DataInput;
import java.io.IOException;
import java.io.*;

/**
 * This is the base class for classes that can marshall themselves.  It
 * implements the infrastructure to package up a class and write it to
 * an output, and then read it back.  It also contains static methods 
 * allowing it to be used with classes that implement Marshallable but
 * do not extend this class.
 */
abstract public class Marshalling implements Marshallable{
  static final String beginDelimiter = "\2BBB\2\n";
  static final String beginDelimiterMatch = "\2BBB\2";
  static final String endDelimiter = "\2XXX\2\n";
  static final String endDelimiterMatch = "\2XXX\2";

  /**
   * This static method marshalls any Marshallable object to the
   * DataOutput out.  It wraps the object's MarshallSelf method by writing
   * the class name and begin and end delimiters.
   */
  public static void marshallSomething(Marshallable obj, DataOutput out) 
    throws IOException {
    out.writeUTF(beginDelimiter);
      
    //then write out the class name
    out.writeUTF(obj.getClass().getName());
      
    obj.MarshallSelf(out);
      
    out.writeUTF(endDelimiter);
  }
  
  /**
   * Returns true if this object can be Marshalled without causing an 
   * Exception.  This is expensive, and should be avoided.
   */
  public boolean canMarshall() {
    DataOutputStream out = new DataOutputStream(new ByteArrayOutputStream());
    try {
      Marshall(out);
    }
    catch (IOException e) {
      return false;
    }
    return true;
  }

  /**
   * Marshall this object by calling marshallSomething with it.
   */
  synchronized public void Marshall(DataOutput out) throws IOException {
    marshallSomething(this, out);
  }

   public static Marshallable Unmarshall(DataInput stream) throws IOException {
     return Unmarshall(stream, null);
   }


  /**
   * This static method takes a DataInput object and extracts a Marshallable
   * from it.  It handles errors by reporting them, but cannot do much about 
   * them.
   */
   public static Marshallable Unmarshall(DataInput in, 
					 ModelObject parent) 
     throws IOException {
    try {
      String beginTag = in.readUTF();
      if (!beginDelimiter.equals(beginTag)) {
	if (beginTag.equals("null"))
	    return null;
	else {
	  System.out.println("Marshalling.Unmarshall: bad begin tag " + beginTag);
	  tryToRecover(in);
	  return null;
	}
      }
    }
    catch (IOException e) {
      System.out.println("threw an exception reading UTF");
//      return null;
    }
    String className = in.readUTF();
    try {
      Class theClass = Class.forName(className);
      Marshallable newObject = (Marshallable)theClass.newInstance();
      if (newObject instanceof Marshalling)
	((Marshalling)newObject).UnmarshallSelf(in, parent);
      else
	newObject.UnmarshallSelf(in);
      String endTag = in.readUTF();
      if (!endDelimiter.equals(endTag)) {
	System.out.println("Marshalling.Unmarshall: bad end tag " + endTag);
	tryToRecover(in);
	return null;
      }
      return newObject;
    } 
    catch (Throwable e) {
      handleException(e, in);
      return null;
    }
  }

  static void handleException(Throwable e, DataInput in) {
    System.out.println(e.getClass() + " in Marshalling.Unmarshall ");
    e.printStackTrace();
    tryToRecover(in);
  }

  static protected void tryToRecover(DataInput in) {
    try {
      String contents = null;
      int recursion = 1;
      do {
	contents = in.readLine();
	System.out.println("Dumping: >>" + contents + "<< size is " + contents.length());
	if (contents.endsWith(beginDelimiterMatch))
	  recursion++;
	else if (contents.endsWith(endDelimiterMatch))
	  recursion--;
      } while (recursion > 0);
    } catch (IOException e) {
      System.out.println("tryToRecover couldn't");
    }
    System.out.println("tryToRecover finished");

  }

  /**
   * Returns the Property determined by the identifier.  Default
   * implementation throws an IllegalArgumentException. When 
   * overriding, invoke inherited::GetProperty for any unknown properties
   * THIS IS NOW OBSOLETE
   */	
  public void GetProperty(String inPropertyID,
			  DataOutput outPropery) 
     throws IllegalArgumentException {
       throw(new IllegalArgumentException("Unknown property in GetProperty"));
     }
	
  /**
   * Returns all important properties (the set of properties needed to
   * clone an object).  Implementations should just call GetProperty for each
   * important property.
   * THIS IS NOW OBSOLETE
   */	
  public void GetImportantProperties(DataOutput outData) {
    return;
  }
	
  /**
   * Sets the Property determined by the identifier, using the data in
   * the stream.  Default implementation throws an IllegalArgumentException
   * When overriding, invoke inherited::SetProperty for any unknown properties
   * THIS IS NOW OBSOLETE
   */	
  public void SetProperty(String inProperty,
			  DataInput inData) throws IllegalArgumentException {
    throw(new IllegalArgumentException("Unknown property in SetPropery"));
  }

  public void UnmarshallSelf(DataInput in, ModelObject parent) 
    throws IOException {
      UnmarshallSelf(in);
    }

}
