/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.io.DataOutput;
import java.io.DataInput;
import java.io.IOException;

/**
 * ModelObject class. ModelObjects are members of an object
 * hierarchy.  They know how to describe themselves in terms of 
 * the hierarchy.  By default, an object describes itself by its position
 * in its parent's list of children, but it can also choose to use a key
 * object to describe itself.
 */
public class ModelObject extends Marshalling implements Specifiable {
  protected ModelObject 	mSuperModel;
  protected SubModels		mSubModels;
  protected static ModelObject mDefaultTopObject = new ModelObject();
  
  
  /**
   * Default Constructor
   */ 
  protected ModelObject() {
  }
  

  /**
   * Constructs a new ModelObject with a given  supermodel and a key.  The 
   * supermodel may be null; this would represent a root object.  The key is
   * is passed to the parent as the desired key for this object.
   */
  public ModelObject(ModelObject inSuperModel, Object key)
    {
      if (inSuperModel != null)
	mSuperModel = inSuperModel;
      else
	mSuperModel = mDefaultTopObject;
      
      if (mSuperModel != null) {
	mSuperModel.AddSubModel(this, key);
      }
      
    }
  
  /**
   * Constructs a new ModelObject with a given  supermodel and no key.
   */
  public ModelObject(ModelObject inSuperModel) {
    this(inSuperModel, null);
  }

  /*
   * Returns the object handling this object's children.  The default is to
   * provide an object which simply identifies a child by its position, but
   * this method can be overridden to provide another object.
   */
  protected SubModels subModels() {
    if (mSubModels == null)
      mSubModels = new SubModelsByPosition();

    return mSubModels;
  }

  /**
   * Return the number of children this object has.
   */	
  public int CountSubModels() {
    return subModels().count();
  }
  
  /**
   * return an Enumeration representing the children this object has.
   */	
  public Enumeration GetSubModels() {
    return subModels().enumerate();
  }
	
  /**
   * Return the top-most object in this hierarchy.
   */
  public ModelObject GetTopModel() {
    if (mSuperModel == mDefaultTopObject)
      return this;
    else
      return mSuperModel.GetTopModel();
  }
  
  /**
   * Return the parent ModelObject.
   */	
  public ModelObject GetSuperModel() {
    return mSuperModel;
  }
  
  /**
   * Set the parent object.  This removes the object from its old parent,
   * and attaches it to the new object.  Call this method
   * with null to remove the subtree formed by this object and its children
   * from the hierarchy 
   */	
  synchronized public void SetSuperModel(ModelObject inObject) {
    if (inObject != mSuperModel) {
      
      if (inObject == this)
	throw new RuntimeException("cannot set supermodel to yourself: " + this);

      if (mSuperModel != null) 
	mSuperModel.RemoveSubModel(this);

      if (inObject != null)
	mSuperModel = inObject;
      else
	mSuperModel = mDefaultTopObject;

      mSuperModel.AddSubModel(this, key());
    }
  }

  /**
   * Adds the given object as a child.  Usually you will
   * not invoke this method; use SetSuperModel on the given object instead.
   */	
  synchronized protected void AddSubModel(ModelObject inObject, Object key) {
    if (subModels().contains(inObject)) {
      // child already exists
      throw new IllegalArgumentException("Submodel already added.");
    }
      
    subModels().add(inObject, key);
  }
  
  /**
   * Removes the given object from the list of children.  Usually you will
   * not invoke this method; use SetSuperModel on the given object instead.
   */	
  synchronized public void RemoveSubModel(ModelObject inObject) {
    subModels().remove(inObject);
  }
  
  /**
   * Determines if the given object is an ancestor of this object.
   */	
  public boolean IsSubModelOf(ModelObject inObject) {
    if (mSuperModel == inObject)
      return true;
    
    if (mSuperModel == null)
      return false;
      
    return mSuperModel.IsSubModelOf(inObject);
  }
  
  /**
   * Returns the child determined by the identifier; returns null
   * if no such child exists.  Does NOT check recursively.
   */	
  public ModelObject GetSubModelByKey(Object key) {
    return subModels().getObject(key);
  }
  
  /**
   * Returns the identifier of the child parameter.  Returns null
   * if no such child exists.  Does NOT check recursively.
   */	
  public Object GetKeyOfSubModel(ModelObject child) {
    return subModels().getKey(child);
  }
  
  /**
   * Make a specifier, starting at the root and write it to the output object.
   */	
  public void MakeSpecifier(DataOutput out) {
    MakeSpecifier(out, null);
  }

  /**
   * Make a specifier, starting at topModel and write it to the output object.
   */	
  public void MakeSpecifier(DataOutput out, ModelObject topModel) {
    try {
      MakeSpecifier(topModel).Marshall(out);
    }
    catch (IOException e) {
      System.out.println("ModelObject.MakeSpecifier: IOException");
    }
  }

  /**
   * Uses recursion to create the Specifier for this object's parent (starting
   * at topModel) and then adds this object's identifier to create its 
   * Specifier.
   */	
  public Specifier AddSpecifier(Specifier inSpecifier, ModelObject topModel) {
    if ((this != topModel) && (mSuperModel != null)) {
      inSpecifier = mSuperModel.AddSpecifier(inSpecifier, topModel);
      return MakeSpecifierSelf(inSpecifier);
    }
    else 
      return inSpecifier;
  }
  
  /**
   * Makes a new specifier for this object starting at the root.
   */	
  public Specifier MakeSpecifier() {
    return MakeSpecifier((ModelObject)null);
  }
  
  /**
   * Makes a new specifier for this object starting at topModel.
   */	
  public Specifier MakeSpecifier(ModelObject topModel) {
    return AddSpecifier(new Specifier(), topModel);
  }
  
  /**
   * Adds the identifier for this object to the end of the Specifier.
   */	
  public Specifier MakeSpecifierSelf(Specifier inSpecifier) {
    if (mSuperModel != null)
      inSpecifier.addElement(mSuperModel.GetKeyOfSubModel(this));
    return inSpecifier;
  }
  
  /**
   * Sets the default root, which is the parent when no other is specified.
   */
  synchronized public static void SetDefaultModel(ModelObject inModel) {
    mDefaultTopObject = inModel;
  }
  
  /**
   * Dump the object to an output object by dumping all its children.
   */	
  public void MarshallSelf(DataOutput out) throws IOException {
    int numSubModels = CountSubModels();
    out.writeInt(numSubModels);
    if (numSubModels > 0) {
      //now do any subObjects
      Enumeration e = GetSubModels();
      while(e.hasMoreElements()) {
	((ModelObject)e.nextElement()).Marshall(out);
      }
    }
  }
  
  /**
   * Read the object's data from an input object.  If there are any
   * children, unmarshall them and set the parent.
   */
  public void UnmarshallSelf(DataInput in) throws IOException {
    int numSubModels = in.readInt();
    for (int i=0; i < numSubModels; i++) {
      ModelObject subModel = (ModelObject)Unmarshall(in);
      if (subModel != null)
	subModel.SetSuperModel(this);
    }    
  }
  
  
  /**
   * Return the Specifiable described by the specifier
   * in relation to myself.  If looking for a ModelObject fails, 
   * return an object which keeps the Specifier as a String.
   */	
  public Specifiable FindSpecified(Specifier inSpecifier) {
//    System.out.println("finding " + inSpecifier);
    return inSpecifier.find(this);
  }
  
  /**
   * This is like FindSpecified except it must return a ModelObject.  If
   * it cannot find one, it returns null.
   */
  public ModelObject FindSpecifiedObject(Specifier inSpecifier) {
    try {
      return (ModelObject)FindSpecified(inSpecifier);
    }
    catch (Exception e) {
      return null;
    }
  }

  /**
   * Find the Specifiable object relative to myself described by the
   *  Specifier which can be created from the given input object.
   */
  public Specifiable FindSpecified(DataInput in) {
    try {
      return FindSpecified((Specifier)Marshalling.Unmarshall(in));
    }
    catch (Exception e) {
      System.out.println("Problem Finding Object " + e);
      return null;
    }
  }

  /**
   * Find the ModelObject relative to myself described by the Specifier 
   * which can be created from the given input object.  If no ModelObject 
   * is found, return null.
   */
  public ModelObject FindSpecifiedObject(DataInput in) {
    try {
      return (ModelObject)FindSpecified(in);
    }
    catch (Exception e) {
      System.out.println("Problem Casting Object to ModelObject " + e);
      return null;
    }
  }
  
  
  /**
   * Find the ModelObject relative to the root described by the Specifier 
   * which can be created from the given input object.  If no ModelObject 
   * is found, return null. Since this does not depend on the object it
   * is called by, this is a static method.
   */
  public static ModelObject FindSpecifiedObjectFromRoot(DataInput in) {
    return mDefaultTopObject.FindSpecifiedObject(in); 
  }
  
 /**
   * Find the ModelObject relative to the root described by the given 
   * Specifier.  If no ModelObject 
   * is found, return null. Since this does not depend on the object it
   * is called by, this is a static method.
   */
  public static ModelObject 
    FindSpecifiedObjectFromRoot(Specifier inSpecifier) {
      return mDefaultTopObject.FindSpecifiedObject(inSpecifier); 
  }
  
  /**
   * Do a depth-first search for a ModelObject which matches matchMe,
   * using useThis as a matching function
   */
  public ModelObject Match(Object matchMe, Matcher useThis) {
    if (useThis.Matches(matchMe, this)) {
      return this;
    }
    else if (CountSubModels() > 0) {
      Enumeration e = GetSubModels();
      while (e.hasMoreElements()) {
	ModelObject m = ((ModelObject)e.nextElement()).Match(matchMe, useThis);
	if (m != null)
	  return m;
	}
    }
    return null;
  }

  /**
   * Do a breadth-first search for a ModelObject which matches matchMe,
   * using useThis as a matching function
   */
  public ModelObject BFMatch(Object matchMe, Matcher useThis) {
    Vector nextMatches = new Vector();
    nextMatches.addElement(this); 
    Enumeration e = nextMatches.elements();
    while (e.hasMoreElements()) {
      ModelObject matched = 
	((ModelObject)e.nextElement()).bfMatch(matchMe, useThis, nextMatches);
      if (matched != null)
	return matched;
    }
    return null;
  }

  /**
   * Private helper function either returns true on a match or adds its
   * children to the list of match candidates.
   */
  private ModelObject bfMatch(Object matchMe, 
			      Matcher useThis, Vector nextMatches) {
    if (useThis.Matches(matchMe, this)) 
      return this;
    else if (CountSubModels() > 0) {
      Enumeration e = GetSubModels();
      while (e.hasMoreElements()) {
	nextMatches.addElement(e.nextElement());
      }
    }
    return null;  
  }	

  public String toString() {
    return super.toString();
  }

  /*
   * Use this to remove a submodel from a hashtable if one is being used
   */
  public Object key() {
    return null;
  }
}	







