/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;
import java.io.*;
import java.net.InetAddress;
import ncsa.habanero.granted.Arbitrator;

/**
 * The NewCollobjectAction is generated by a client when a user adds
 * a new Collobject to a session.  It is sent to the Arbitrator, which,
 * because the NewCollobjectAction implements the ArbitratorAction interface,
 * calls actAtArbitrator on it.
*/
public class NewCollobjectAction extends Action implements ArbitratorAction {
  protected String collobName;
  protected String arbitratorName;

  /**
   * Initialize the NewCollobjectAction with the name of the Collobject it
   * should initiate and the type of Arbitrator the Collobject should have.
   */
  public NewCollobjectAction(String collobName, String arbitratorName) {
    super();
    this.collobName = collobName;
    if (arbitratorName == null)
      arbitratorName = "ncsa.habanero.CentralArbitrator";
    this.arbitratorName = arbitratorName;
  }
  
  /**
   * Initialize the NewCollobjectAction with the name of the Collobject it
   * should initiate and a CentralArbitrator.
   */
  public NewCollobjectAction(String collobName) {
    this(collobName, null);
  }
  
  /**
   * This should only be called if the next call will be a marshall()
   */
  public NewCollobjectAction() {
    super();
  }
  
  /**
   * Since this particular action is supposed to act only at the Arbitrator,
   * this method should not be called.
   */
  public void act(Targetable target) {
  }

  /**
   * Write the Collobject name to the output.
   */
  public void MarshallSelf(DataOutput out) throws IOException {
    super.MarshallSelf(out);
    out.writeUTF(collobName);
    out.writeUTF(arbitratorName);
  }

  /**
   * Read the Collobject name from the input.
   */
  public void UnmarshallSelf(DataInput in) throws IOException {
    super.UnmarshallSelf(in);
    collobName = in.readUTF();
    arbitratorName = in.readUTF();
  }
  
  /**
   * At the Arbitrator, we make a Collobject of the specified type with
   * an Arbitrator of the specified type.  Since this is on a server, the
   * Arbitrator must be of type CentralArbitrator.
   * This is still kind of a HACK - we also need to specify whether
   * a new Router should be created.  Then we create a new action request,
   * with an Action which asks the clients to create the same Collobject,
   * and tells their DistributedArbitrators where to connect to.
   */
  public void actAtArbitrator(Arbitrator arb, Specifiable target) {
    if (arb.session() == null)
      return;

    if (!arb.session().isShared()) {
      Action newAction = new AddCollobjectAction(collobName, null,
						 1, arb, originator);
      newAction.act((Targetable)target);
      return;
    }

    try {
      Class theClass = Class.forName(arbitratorName);
      CentralArbitrator collarb = (CentralArbitrator)theClass.newInstance();
      collarb.init(arb, null);
      new Collobject((Sharable)target, null, collarb, originator);
      try {
	Action newAction = new AddCollobjectAction(collobName, 
						   InetAddress.getLocalHost(),
						   collarb.port(),
						   arb,
						   originator);
	new StandardActionRequest(newAction, target);
      }
      catch (java.net.UnknownHostException uhe) {
	System.out.println("UnknownHostException in actAtArbitrator ");
      }
    }
    catch (ClassCastException e) {
      System.out.println("ClassCastException in actAtArbitrator ");
      System.out.println(arbitratorName + " must be a subclass of CentralArbitrator" );
    }
    catch (ClassNotFoundException e) {
      System.out.println("ClassNotFoundException in actAtArbitrator ");
      System.out.println("Tried to find a " + arbitratorName);
    }
    catch(InstantiationException e) {
      System.out.println("InstantiationException in actAtArbitrator ");
      System.out.println("Couldn't instantiate a " + arbitratorName);
    }
    catch(IllegalAccessException e) {
      System.out.println("IllegalAccessException in actAtArbitrator ");
      System.out.println(arbitratorName + " must be public and must have a public default constructor");
    }
  }

  public boolean sendToEveryone() {
    return false;
  }

}

