/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero;
import java.util.*;
import java.io.*;
import ncsa.habanero.granted.ActionGranted;
import ncsa.habanero.granted.Serializer;
import ncsa.env.control.PacSession;

/**
 * A Router sends and receives ActionGranteds. It is responsible for
 * sending an ActionGranted to other routers when given one by an Arbitrator,
 * and for receiving ActionGranteds from other routers and passing them to 
 * a serializer.  The Router implements the Observer interface, allowing
 * it to be updated whenever a participant is added or removed from a 
 * session.
 */
abstract public class Router implements java.util.Observer {
  protected Sharable owner;
  protected int frozen = 0;
  protected boolean stopped = false;

  /**
   * Upon construction, a router registers itself with a Session, allowing
   * it to be updated whenever a participant is added or removed from a 
   * session.
   *@param owner the Sharable which owns the arbitrator I
   * am routing for
  */
  Router(Sharable owner) {
    this.owner = owner;
    if (owner != null)
      owner.getSession().register(this);
  }
  
  /**
   * Sometimes the owner cannot be set in the constructor, and is set by
   * this method instead.  This method only works if the owner is null.
   */
  void set_owner(Sharable owner) {
    if ((this.owner == null) && (owner != null)){
      this.owner = owner;
      owner.getSession().register(this);
    }
  }

  /**
   * Send the ActionGranted to every participant. This should be overridden.
   *@param request
  */
  abstract protected void send_to_everyone(ActionGranted request);

  /**
   * Return an Enumeration representing the participants.  This should 
   * be overridden.
  */
  abstract protected Enumeration participants();

  /** 
   * This is called when input is made available.  First, it checks the
   * frozen instance variable, and if it is supposed to be frozen, goes 
   * into wait.  When it is not frozen, it unmarshalls the ActionGranted
   * and sends it to its Serializer.  If it cannot unmarshall properly,
   * it prints an error and exits, because someone really ought to fix the
   * problem.  This routine cannot be synchronized, because it is always
   * blocking in Unmarshall, but it calls checkTemperature to block
   * if input should be frozen.
   */
  protected boolean handleInput(DataInput in)  {
    checkTemperature();
    try {
      ActionGranted granted = (ActionGranted)Marshalling.Unmarshall(in);
      if (granted != null)
	granted.serializer().incoming(granted);
      else
	System.out.println("no ActionGranted made in Router.handleInput()");
    } 
    catch (java.io.IOException e) {
      System.out.println("IOException in Router.handleInput()");
      return false;
    }
    return true;
  }
  
  /**
   * This routine blocks while the variable frozen is > 0.
   */
  synchronized protected void checkTemperature() {
    while (frozen > 0) {
      try {
	wait();
      }
      catch (java.lang.InterruptedException e) {
	System.out.println("Thread interrupted in wait state");
	System.out.println("Ignoring exception : we should still wait");
      }
    }
  }

  /**
   * This is used when dumping state to ensure the state doesn't change 
   * while we are dumping.  It simply flags that the router should be frozen.
   * The next input will not be handled until frozen is 0.
   */
  synchronized void freezeInput() {
    frozen += 1;
    System.out.println("freezeInput: frozen for " + this + " is now " + frozen);
  }

  /**
   * Decrements frozen by one and wakes up any waiting threads.  When frozen
   * hits 0, the Router will start handling actions again.
   */
  synchronized void thawInput() {
    if (frozen > 0)
      frozen -= 1;
    System.out.println("thawInput: frozen for " + this + " is now " + frozen);
    notifyAll();
  }

  public synchronized void stop() {
    stopped = true;
    owner.getSession().unregister(this);
    owner = null;
  }
      
  /**
   * This method is called by o, which is a ParticipantManager, when arg,
   * which is a SessionParticipant is added to or removed from the Session.
   * It should be overridden.
   *@param o the ParticipantManager for the session
   *@param arg the SessionParticipant whose status has changed.
  */
  public void update(Observable o, Object arg) {
  }
}

