/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero.granted;
import ncsa.habanero.*;
import ncsa.habanero.tickets.*;
import java.io.DataOutput;
import java.io.DataInput;
import java.io.IOException;

/**
 * The Arbitrator is the abstract superclass for all floor control objects.  
 * The basic idea is that an Arbitrator is sent an ActionRequest.  After some
 * sort of checking or negotiation, the Arbitrator grants the request, 
 * creates an ActionGranted, and sends that to a router.  Arbitrators live in
 * a hierarchy, being subclasses of ModelObject, so each one knows its parent.
 */
abstract public class Arbitrator extends ModelObject {
  protected Sharable owner;
  /**
   * The constructor sets the parent arbitrator
   * @param parent the parent object for the arbitrator
  */
  protected Arbitrator(ModelObject parent) {
    super(parent);
  }

  /**
   * This routine should be overridden to do something with the
   * request.  As a result of this method,  an ActionGranted should be 
   * sent to the router.
   */
  abstract public synchronized boolean arbitrate(ActionRequest request);

    /**
	 *
     *  getLock
     *
     *  Requires:  lkdNew
     *  Modifies:  itself
     *  Effects:   returns a lock built from the lock descriptor if it
	 *				is OK to do so
	 *
     */

  abstract public Key getLock (Lock lkdNew);
  abstract public Key getLock (Lock lkdNew, Key oldKey);
  public synchronized Key blockingGetLock (Lock ldNew, Key oldKey) {
    return getLock(ldNew, oldKey);
  }

    /**
     *
     *  releaseLock
     *
     *  Requires:  lkFinished - the key of the lock that is no longer needed
     *  Modifies:  itself
     *  Effects:   releases the lock
	 *
     */

    abstract public void releaseLock (Key lkFinished);
  abstract public synchronized void stop();

  /**
   * Returns a new ActionGranted for the given request.  This is
   * here because the ActionGranted constructor is protected, and
   * we want new subclasses of the arbitrator to be able to grant
   * requests even if they are not in the same package.
   * @param request the ActionRequest we want to grant
  */
  protected ActionGranted grant_request(ActionRequest request) {
    return new ActionGranted(request, 
			     ticket_granter().grant_ticket(request.target()),
			     this);
  }

  /**
   * This should only be called with an Action that is generated by this
   * Arbitrator, not with any Action that was received as an ActionRequest.
   */
  protected ActionGranted grantNewAction(Action action, Specifiable target) {
    return new ActionGranted(action,
			     ticket_granter().grant_ticket(target),
			     this);
  }

  /** 
   * Return the TicketGranter used by this Arbitrator.  This would be 
   * protected, but because of the language, we cannot use delegation with
   * protected methods, and we need DynamicArbitrator to delegate this.
   */
  abstract public TicketGranter ticket_granter();

  /** 
   * Return the Serializer used by this Arbitrator.  This would be 
   * protected, but because of the language, we cannot use delegation with
   * protected methods, and we need DynamicArbitrator to delegate this.
   */
  abstract public Serializer serializer();

  /**
   * Set the Sharable which owns this object.  This is used to rectify a
   * chicken-egg problem when constructing ModelObjects and Arbitrators.
   */
  public synchronized void set_owner(Sharable owner) {
    this.owner = owner;
  }

  public Sharable getOwner() {
    return owner;
  }

  /**
   * Return the Session which this object is in.
   */
  public Session session() {
    try {
      return ((Arbitrator)GetSuperModel()).session();
    }
    catch (ClassCastException e) {
      System.out.print("supermodel is not an Arbitrator, it is a ");
      System.out.println(GetSuperModel().getClass().getName());
      return null;
    }
  }

  /**
   * Return the router this Arbitrator uses.  The default is to use
   * my parent's router.  For this to work, there must actually be a router 
   * at the top-level Arbitrator.
   */
  public Router router() {
    try {
      return ((Arbitrator)GetSuperModel()).router();
    }
    catch (ClassCastException e) {
      System.out.print("supermodel is not an Arbitrator, it is a ");
      System.out.println(GetSuperModel().getClass().getName());
      return null;
    }
  }

  /**
   * Write some subclass-specific information to the output object.
   */
  public void writeInfo(DataOutput out) {
    // nothin to write 
  }

  /**
   * Write out the Specifier for my owner.
   */
  public void MarshallSelf(DataOutput outStream) throws IOException {
    owner.MakeSpecifier(outStream, SessionManager.manager());
  }

  /**
   * Find my owner, given its Specifier.
   */
  public void UnmarshallSelf(DataInput inStream) throws IOException {
    SessionManager.manager().FindSpecifiedObject(inStream);
  }

  public boolean isDistributed() {
    return true;
  }

  /**
   * Makes a new specifier for this object starting at topModel.
   */	
  public Specifier MakeSpecifier(ModelObject topModel) {
    return getOwner().AddSpecifier(new ArbitratorSpecifier(), topModel);
  }
  

}

