/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero.granted; 

import java.util.*;
import java.io.*;
import ncsa.habanero.*;
import ncsa.habanero.tickets.Ticket;

/**
 * The Serializer class makes sure that each action for which it is 
 * responsible (at present, that means all the actions for a particular 
 * collobject) is executed in order, and only once.  To that end, it 
 * queues out of order ActionGranteds, and throws away duplicate ones.
 * There is conceptually one serializer for each TicketGranter, and thus
 * one for each Arbitrator.  
*/
public class Serializer extends Marshalling {
  protected int last_one_done = 0;
  protected Hashtable pending = new Hashtable();
  protected boolean queueNewActions = false;
  protected Arbitrator arbi;
  protected static Vector dumpInfo = null;

  /**
   * Default constructor.
   */
  public Serializer() {
  }

  public Serializer(Arbitrator arb) {
    arbi = arb;
  }
  
  /**
   * This is used to tell the Serializer to queue actions rather than 
   * allowing them to be executed.  It can be used to "freeze" an Arbitrator.
   */
  public void startQueueing() {
    queueNewActions = true;
  }

  /**
   * Stop queueing actions, and execute any in the queue.
   */
  public void endQueueing() {
    queueNewActions = false;
    checkPendingActions();
  }

  /**
   * Tells whether the ActionGranted's ticket number matches the Serializer's
   * idea of the next number to execute.  If queueing is enables, it returns
   * Ticket.NOT_YET, which results in queueing the ActionGranted.
   */
  protected int matches(ActionGranted request) {
//    System.out.println("Serializer " + this + " matching a request");
    if (queueNewActions) 
      return Ticket.NOT_YET;
    else 
      return request.ticket().matches(last_one_done+1);
  }

  
  /**
   * This method is called by the router.  It checks if the 
   * Action represented by the ActionGranted can be executed, and
   * if so asks the ActionGranted to do so.  It queues out of order
   * ActionGranteds, and throws away duplicate ones. 
   * @param request    the action that has been granted
  */
  public void incoming(ActionGranted request) {
    switch (matches(request)) {
    case Ticket.MATCHES:
      request.OK();
      ++last_one_done;
      checkPendingActions();
      break;
    case Ticket.NOT_YET:
      queueRequest(request);
      break;
    case Ticket.ALREADY_SEEN:
      // throw it away
      break;
    }
  }
  
  /**
   * After an action is received, this checks if the next one is
   * in the queue.  If so, it calls incoming with it.  Note that
   * this is then recursively called.
   */
  protected void checkPendingActions() {
    if (!pending.isEmpty()) {
      ActionGranted nextAction = 
	(ActionGranted)pending.remove(new Integer(last_one_done+1));
      if (nextAction != null) {
	System.out.println("trying action agaain: " + nextAction.ticket().serialNumber());
	incoming(nextAction);
      }
    }
  }   

  /**
   * This is called to put an early ActionGranted in our queue,
   * which is actually a hashtable indexed by ticketNumber
  */
  protected void queueRequest(ActionGranted request) {
    pending.put(new Integer(request.ticket().serialNumber()), request);
  }

  /**
   * Marshall myself by writing out the ticket number of the last 
   * ActionGranted I matched.
   */
  public void MarshallSelf(DataOutput outStream) throws IOException {
    outStream.writeInt(last_one_done);
  }

 /**
   * Unmarshall myself by reading the ticket number of the last 
   * ActionGranted I matched.
   */
  public void UnmarshallSelf(DataInput inStream, 
			     ModelObject arbitrator) throws IOException {
    arbi = (Arbitrator)arbitrator;
    last_one_done = inStream.readInt();
    System.out.println("Serializer unmarshalled " + last_one_done);
    dumpRecord();
  }

  public void UnmarshallSelf(DataInput inStream) throws IOException {
    System.out.println("Serializer.UnmarshallSelf() should not be called with a single argument");
    UnmarshallSelf(inStream, null);
  }
  
  /**
   * This is used to start recording info on the serial number a Serializer
   * is initialized with.
   */
  static public void startRecording() {
    dumpInfo = new Vector();
  }

  protected synchronized void dumpRecord() {
    dumpInfo.addElement(new DumpInfoRecord(arbi, last_one_done));
  }    

  static public void writeRecords(DataOutput out) {
    int dataSize = (dumpInfo == null) ? 0 : dumpInfo.size();
    System.out.println("writing " + dataSize + " Specifier records");
    try {
      out.writeInt(dataSize);
      for (int index = 0; index < dataSize; index++) {
	((DumpInfoRecord)dumpInfo.elementAt(index)).write(out);
      }
    } catch (java.io.IOException e) {
      System.out.println("could not write records because " + e.getMessage());
    }

  }

  static public Vector readRecords(DataInput in) {
    try {
      int dataSize = in.readInt();
      System.out.println("reading " + dataSize + " Specifier records");

      if (dataSize == 0)
	return null;

      Vector result = new Vector(dataSize);
      for (int index = 0; index < dataSize; ++index) {
	result.addElement(new DumpInfoRecord(in));
      }
      return result;
    } catch (java.io.IOException e) {
      System.out.println("could not read records because " + e.getMessage());
      return null;
    }
  }
}
































