/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *  
 * info@havefun.ncsa.uiuc.edu
 *  
 * Copyright 1995-1996, Board of Trustees of the University of Illinois, 
 * All Rights Reserved
 *  
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *  
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *  
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *  
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *  
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *  
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 * 
 *
 * ----
 * 
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * 
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 * 
 */
package ncsa.habanero.tickets;
import ncsa.habanero.*;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
/**
 * The ticket is created (only) by a TicketGranter, and represents permission
 * for an associated Action to be executed on a particular target in an
 * order represented by its serial number.  Out of order tickets are 
 * queued by the Serializer until previous tickets arrive.  The Serializer
 * also passes the associated Action the target upon which to execute.  The
 * Ticket is opaque, allowing only a matches operation.
*/
public class Ticket extends Marshalling {
  public static final int MATCHES = 0;
  public static final int ALREADY_SEEN = 1;
  public static final int NOT_YET = 2;
  public static final int WRONG_TARGET = 3;
  
  private int serialNumber = -1;
  private Specifiable target;

  /**
   * Construct a Ticket with an order and a target.  This is a default
   * access constructor, meant to allow only TicketGranters to use it
   * @param num The serial number
   * @param tgt The target
   */
  Ticket(int num, Specifiable tgt) {
    serialNumber = num;
    target = tgt;
  }

  /** 
   * Default constructor, used for Marshalling.
   */
  public Ticket() {
  }

  /**
   * Return the target.  In some cases the object stored as the target is
   * not Targetable; an Exception will be thrown in such a case.
   */
  Targetable target() {
    return (Targetable)target;
  }
  
  /**
   * This returns the ticket's serial number.  I'm making this public 
   * against my better judgement, but it's
   * needed by Serializer, which cannot be in this package.
   */
  public int serialNumber() {
    return serialNumber;
  }
  
  /**
   * We do not really want the target to go outside the ticket, so
   * we have the ticket make the call to the action.
   * @param action The action to perform
   */
  public void act(Action action) {
    target().performAction(action);
  }
  
  /**
   * Returns different values based on whether the serial number matches
   * the number passed as a parameter.
   */
  public int matches(int number) {
    if (serialNumber == number) 
      return MATCHES;
    else if (serialNumber > number) {
      System.out.println("saving:");
      System.out.print("number is: " + number);
      System.out.println(" serial is: " + serialNumber);
      return NOT_YET;
    }
    else {
      System.out.println("throwing away:");
      System.out.print("number is: " + number);
      System.out.println(" serial is: " + serialNumber);
      return ALREADY_SEEN;
    }
  }

  /**
   * Read the serial number and the find the target from its Specifier.
   */
  public void UnmarshallSelf(DataInput stream) throws IOException {
    serialNumber = stream.readInt();
    target = SessionManager.manager().FindSpecified(stream);
    if (target == null)
      throw new IOException("Ticket target is null");
  }
  
  /**
   * Write the serial number and the Specifier of the target.
   */
  public void MarshallSelf(DataOutput stream) throws IOException {
    stream.writeInt(serialNumber);
    target.MakeSpecifier(stream, SessionManager.manager());
  }

}

