/*
 * NCSA Habanero Project
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign, IL 61820
 *
 * info@habanero.ncsa.uiuc.edu
 *
 * Copyright 1995-1996, Board of Trustees of the University of Illinois,
 * All Rights Reserved
 *
 * NCSA Habanero software, binary, Java-bytecode, and source (hereafter,
 * Software) is copyrighted by The Board of Trustees of the University of
 * Illinois (UI), and ownership remains with the UI.
 *
 * The UI grants you (hereafter, Licensee) a license to use the Software for
 * academic, research and internal business purposes only, without a fee.
 * Licensee may distribute the binary and source code (if released) to third
 * parties provided that the copyright notice and this statement appears on
 * all copies and that no charge is associated with such copies.
 *
 * Licensee may make derivative works.  However, if Licensee distributes any
 * derivative work based on or derived from the Software, then Licensee will
 * (1) notify NCSA regarding its distributing of the derivative work, and (2)
 * clearly notify users that such derivative work is a modified version and
 * not the original NCSA Habanero distributed by the UI.
 *
 * Any Licensee wishing to make commercial use of the Software should contact
 * the UI, c/o NCSA, to negotiate an appropriate license for such commercial
 * use.  Commercial use included (1) integration of all or part of the
 * Software into a product for sale or license by or on behalf of Licensee to
 * third parties, or (2) distribution of the Software to third parties that
 * need it to utilize a commercial product sold or licensed by or on behalf of
 * Licensee.
 *
 * UI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
 * PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.  THE
 * UI SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY THE USERS OF THIS
 * SOFTWARE.
 *
 * By using or copying this Software, Licensee agrees to abide by the
 * copyright law and all other applicable laws of the U.S. including, but not
 * limited to, export control laws, and the terms of this license.  UI shall
 * have the right to terminate this license immediately by written notice upon
 * Licensee's breach of, or non-compliance with, any of its terms.  Licensee
 * may be held legally responsible for any copyright infringement that is
 * caused or encouraged by Licensee's failure to abide by the terms of this
 * license.
 *
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * The Software and/or user documentation are provided with RESTRICTED AND
 * LIMITED RIGHTS.  Use, duplication or disclosure by or to the United States
 * government shall be subject to restrictions as set forth in subparagraph
 * (c)(1)(ii) of Defense Federal Acquisition Regulations Supplement (DFARS)
 * Section 252.227-7013 for Department of Defense contracts and as set forth
 * in Federal Acquisition Regulations (FAR) Section 52.227-19 for civilian
 * agency contracts or any successor regulations.  Contractor/Manufacturer is
 * The Board of Trustees of the University of Illinois, Urbana, Illinois
 * 61801.
 */
package WhiteBoard;

import java.awt.*;
import java.io.*;
import java.util.*;

/**
 * A class that produces a image button component.
 *
 * @version 0.1 01/05/96
 * @author  Stephen R. Pietrowicz
 */
public class ImageButton extends Canvas {

	public final static int NORMAL = 100;
	public final static int SELECTED = 101;
	public final static int DISABLED = 102;
	public final static int HIGHLIGHT = 103;

	private int button_state = 0;

	/**
	 * The image of the button when it has not been touched.
	 */
	Image ib_normal = null;

	/**
	 * The image of the button when it has been selected.
	 */
	Image ib_selected = null;

	/**
	 * The image of the button when it has disabled.
	 */
	Image ib_disabled = null;

	/**
	 * The image of the button when the cursor moves over it.
	 */
	Image ib_highlight = null;

	private Image current = null;

	public boolean retain_state = false;

	/**
	 * Constructs a button with an image, but no alternate images
	 * @param normal the file containing the normal button image to use
	 */
	public ImageButton(String normal) {
		this(normal, null, null, null);
	}

	/**
	 * Constructs a button with an image, and an alternate image
	 * when the button has been pressed.
	 * @param normal the file containing the normal button image
	 * @param selected the file containing the selected button image
	 */
	public ImageButton(String normal, String selected) {
		this(normal, selected, null, null);
	}

	/**
	 * Constructs a button with an image, an alternate image
	 * when the button has been pressed, and a image to use when
	 * the button has been disabled.
	 * @param normal the file containing the normal button image
	 * @param selected the file containing the selected button image
	 * @param disabled the file containing the disabled button image
	 */
	public ImageButton(String normal, String selected, String disabled) {
		this(normal, selected, disabled, null);
	}

	/**
	 * Constructs a button with an image, an alternate image
	 * when the button has been pressed, a image that is displayed when
	 * the button has been disabled, and an image that is displayed when 
	 * the cursor is over the button.
	 * @param normal the file containing the normal button image
	 * @param selected the file containing the selected button image
	 * @param disabled the file containing the disabled button image
	 * @param hightlight the file containing the hightlight button image
	 */
	public ImageButton(String normal, String selected, String disabled, String highlight) {
		Image img = null;

		MediaTracker tracker = new MediaTracker(this);
		
		if (normal != null) {
			img = Toolkit.getDefaultToolkit().getImage(normal);
			ib_normal = img;
			tracker.addImage(ib_normal, 0);
		}

		if (selected != null) {
			img = Toolkit.getDefaultToolkit().getImage(selected);
			ib_selected = img;
			tracker.addImage(ib_selected, 0);
		}

		if (disabled != null) {
			img = Toolkit.getDefaultToolkit().getImage(disabled);
			ib_disabled = img;
			tracker.addImage(ib_disabled, 0);
		}

		if (highlight != null) {
			img = Toolkit.getDefaultToolkit().getImage(highlight);
			ib_highlight = img;
			tracker.addImage(ib_highlight, 0);
		}

		try {
			tracker.waitForID(0);
		} catch (InterruptedException e) {
			return;
		}
			
		int width = maxWidth(ib_normal, ib_selected, ib_disabled);
		int height = maxHeight(ib_normal, ib_selected, ib_disabled);

		setState(NORMAL);
		resize(width, height);
	}

	private int maxWidth(Image normal, Image selected, Image disabled) {
		int width = 0;
		
		if (ib_normal != null) 
			width = ib_normal.getWidth(this);
		if (ib_selected != null) 
			width = Math.max(width, ib_selected.getWidth(this));
		if (ib_disabled != null) 
			width = Math.max(width, ib_disabled.getWidth(this));

		return width;
	}

	private int maxHeight(Image normal, Image selected, Image disabled) {
		int height = 0;
		
		if (ib_normal != null) 
			height = ib_normal.getHeight(this);
		if (ib_selected != null) 
			height = Math.max(height, ib_selected.getHeight(this));
		if (ib_disabled != null) 
			height = Math.max(height, ib_disabled.getHeight(this));

		return height;
	}

	public void paint(Graphics g) {
		if (current != null)
			g.drawImage(current, 0, 0, this);
	}

	public boolean mouseDown(Event e, int x, int y) {
		if (retain_state) {
			if (getState() == SELECTED) {
				setState(NORMAL);
			} else {
				setState(SELECTED);
			}
		} else {
			setState(SELECTED);
		}

		paint(this.getGraphics());

		return true;
	}

	public boolean mouseUp(Event e, int x, int y) {

		if (retain_state) 
			return true;

		if (ib_normal != null) {
			setState(NORMAL);
			paint(this.getGraphics());
		}

		action();


		return true;
	}

	public void setState(int state) {
		switch (state) {
			case NORMAL:
				current = ib_normal;
				break;
			case SELECTED:
				current = ib_selected;
				break;
			case DISABLED:
				current = ib_disabled;
				break;
			case HIGHLIGHT:
				current = ib_highlight;
				break;
			default:
				return;
		}
		button_state = state;
		repaint();
	}

	public int getState() {
		return button_state;
	}

	/**
	 * The action the button will initiate.  Returns true if
	 * successful, false otherwise.
	 */
	public boolean action() {
		return true;
	}
}
